"""URLs for the ``event_rsvp`` app."""
from django.conf.urls.defaults import patterns, url

from . import views


urlpatterns = patterns(
    '',
    url(r'^(?P<pk>\d+)/$',
        views.EnquirySubmitView.as_view(),
        name='enquiry_detail'),

    url(r'^$',
        views.EnquiryListView.as_view(),
        name='enquiry_list'),
)



"""Factories for the ``enquiry`` app."""
from django.utils import timezone

from django_libs.tests.factories import UserFactory
import factory

from enquiry import models


class EnquiryFactory(factory.Factory):
    """Factory for the ``Enquiry`` model."""
    FACTORY_FOR = models.Enquiry

    created_by = factory.SubFactory(UserFactory)
    start_date = factory.LazyAttribute(
        lambda a: timezone.now() - timezone.timedelta(days=1))
    end_date = factory.LazyAttribute(
        lambda a: timezone.now() + timezone.timedelta(days=7))
    is_published = True


class EnquiryTransFactoryBase(factory.Factory):
    """Base factory for factories for ``EnquiryTrans`` models."""
    FACTORY_FOR = models.EnquiryTrans

    enquiry = factory.SubFactory(EnquiryFactory)


class EnquiryTransENFactory(EnquiryTransFactoryBase):
    """Factory for english ``EnquiryTrans`` objects."""
    question = 'A question?'
    language = 'en'


class EnquiryTransDEFactory(EnquiryTransFactoryBase):
    """Factory for german ``EnquiryTrans`` objects."""
    question = 'Eine Frage?'
    language = 'de'


class AnswerFactory(factory.Factory):
    """Factory for the ``Answer`` model."""
    FACTORY_FOR = models.Answer

    enquiry = factory.SubFactory(EnquiryFactory)


class AnswerTransFactoryBase(factory.Factory):
    """Base factory for factories for ``EnquiryTrans`` models."""
    FACTORY_FOR = models.AnswerTrans

    answer = factory.SubFactory(AnswerFactory)


class AnswerTransENFactory(AnswerTransFactoryBase):
    """Factory for english ``AnswerTrans`` objects."""
    text = 'An answer'
    language = 'en'


class AnswerTransDEFactory(AnswerTransFactoryBase):
    """Factory for german ``AnswerTrans`` objects."""
    text = 'Eine Antwort'
    language = 'de'


class VoteFactory(factory.Factory):
    """Factory for the ``Answer`` model."""
    FACTORY_FOR = models.Vote

    answer = factory.SubFactory(AnswerFactory)
    user = factory.SubFactory(UserFactory)



"""Tests for models of the ``enquiry``` application."""
from django.test import TestCase
from django.utils import timezone

from enquiry.tests.factories import (
    AnswerFactory,
    EnquiryFactory,
    EnquiryTransDEFactory,
    EnquiryTransENFactory,
    VoteFactory,
)


class EnquiryTestCase(TestCase):
    """Tests for the ``Enquiry`` model class."""
    longMessage = True

    def setUp(self):
        self.obj = EnquiryFactory()

    def test_model(self):
        self.assertTrue(self.obj.pk)

    def test_translation(self):
        self.assertFalse(self.obj.get_translation())
        EnquiryTransDEFactory(enquiry=self.obj)
        self.assertEqual(self.obj.get_translation().question, 'Eine Frage?')
        EnquiryTransENFactory(enquiry=self.obj)
        self.assertEqual(self.obj.get_translation().question, 'A question?')

    def test_is_active(self):
        self.assertTrue(self.obj.is_active(), msg=(
            'Should be True if the campaign is up-to-date.'))
        upcoming_event = EnquiryFactory(
            start_date=timezone.now() + timezone.timedelta(days=1))
        self.assertFalse(upcoming_event.is_active(), msg=(
            'Should be False if the campaign has not started.'))


class AnswerTestCase(TestCase):
    """Tests for the ``Answer`` model class."""
    longMessage = True

    def setUp(self):
        self.obj = AnswerFactory()

    def test_model(self):
        self.assertTrue(self.obj.pk)

    def test_get_vote_count(self):
        self.assertEqual(self.obj.get_vote_count(), 0)
        VoteFactory(answer=self.obj)
        self.assertEqual(self.obj.get_vote_count(), 1)


class VoteTestCase(TestCase):
    """Tests for the ``Vote`` model class."""
    longMessage = True

    def test_model(self):
        obj = VoteFactory()
        self.assertTrue(obj.pk)



"""Tests for tags of the ``enquiry``` application."""
from django.contrib.auth.models import AnonymousUser
from django.contrib.sessions.middleware import SessionMiddleware
from django.template.context import RequestContext
from django.test import TestCase
from django.test.client import RequestFactory

from ..templatetags.enquiry_tags import (
    get_answers,
    get_current_poll,
    render_current_poll,
)
from enquiry.tests.factories import (
    AnswerTransENFactory,
    EnquiryTransENFactory,
    VoteFactory,
)


class GetAnswersTestCase(TestCase):
    """Tests for the ``get_answers`` tag."""
    longMessage = True

    def test_tag(self):
        enquiry = EnquiryTransENFactory()
        answer1 = AnswerTransENFactory(answer__enquiry=enquiry.enquiry)
        answer2 = AnswerTransENFactory(answer__enquiry=enquiry.enquiry)
        vote = VoteFactory(answer=answer2.answer)
        result = get_answers(enquiry.enquiry)
        self.assertEqual(result[0], answer1, msg=(
            'When called without sort it should return the answers ordered by'
            ' ID (Django default)'))
        result = get_answers(enquiry.enquiry, sort=True)
        self.assertEqual(result[0], answer2, msg=(
            'When called without sort it should return the answers ordered by'
            ' vote count'))


class RenderCurrentPollTestCase(TestCase):
    """Tests for the ``render_current_poll`` tag."""
    longMessage = True

    def test_tag(self):
        # create context mock
        request = RequestFactory().get('/')
        request.user = AnonymousUser()
        SessionMiddleware().process_request(request)
        request.session.save()
        context = RequestContext(request)

        # Returns empty string if there's no current poll
        self.assertEqual(render_current_poll(context), '')

        # Returns current_poll.html
        self.vote = VoteFactory()
        enquiry_translated = EnquiryTransENFactory(
            enquiry=self.vote.answer.enquiry)
        AnswerTransENFactory(answer=self.vote.answer)
        self.assertEqual(render_current_poll(context), {
            'has_voted': False,
            'enquiry': enquiry_translated.enquiry,
        })


class GetCurrentPollTestCase(TestCase):
    """Tests for the ``get_current_poll`` tag."""
    longMessage = True

    def test_tag(self):
        # Returns None if there's no current poll
        self.assertIsNone(get_current_poll())

        # Returns current tag
        self.vote = VoteFactory()
        enquiry_translated = EnquiryTransENFactory(
            enquiry=self.vote.answer.enquiry)
        AnswerTransENFactory(answer=self.vote.answer)
        self.assertEqual(get_current_poll(), enquiry_translated.enquiry)



"""Views for the ``event_rsvp`` app."""
from django.core.urlresolvers import reverse
from django.http import Http404, HttpResponseRedirect
from django.views.generic import DetailView, ListView

from .models import Answer, Enquiry, Vote


class EnquiryListView(ListView):
    """List view for the Enquiry model."""

    def get_queryset(self):
        return Enquiry.objects.filter(is_published=True)


class EnquirySubmitView(DetailView):
    """Display an enquiry and gather votes."""
    model = Enquiry

    def dispatch(self, request, *args, **kwargs):
        """Get the relevant enquiry and check if this vote can be anonymous."""
        self.kwargs = kwargs
        self.object = self.get_object()
        if request.method == 'POST':
            if (not self.object.allow_anonymous
                    and not request.user.is_authenticated()):
                raise Http404
            try:
                answer_id = int(request.POST.get('answer_pk'))
            except ValueError:
                raise Http404
            try:
                self.answer = Answer.objects.get(pk=answer_id)
            except Answer.DoesNotExist:
                raise Http404
            if self.answer.enquiry != self.object:
                raise Http404
            if not self.object.has_voted(request):
                if not request.user.is_authenticated():
                    Vote.objects.create(
                        answer=self.answer,
                        session_key=request.session.session_key)
                else:
                    Vote.objects.create(
                        user=request.user, answer=self.answer,
                        session_key=request.session.session_key)
            return HttpResponseRedirect(reverse(
                'enquiry_detail', kwargs={'pk': self.object.pk}))
        return super(EnquirySubmitView, self).dispatch(request, *args,
                                                       **kwargs)

    def get_context_data(self, *args, **kwargs):
        context = super(EnquirySubmitView, self).get_context_data(*args,
                                                                  **kwargs)
        if self.object.has_voted(self.request):
            context.update({'has_voted': True})
        return context



"""Models for the ``enquiry`` app."""
from django.conf import settings
from django.db import models
from django.utils import timezone
from django.utils.translation import get_language
from django.utils.translation import ugettext_lazy as _

from simple_translation.utils import get_translation_queryset


class TransModelMixin(object):
    """Mixin that returns the translation object for a given parent model."""
    def get_translation(self):
        lang = get_language()
        queryset = get_translation_queryset(self).filter(language=lang)
        if queryset:
            return queryset[0]
        # If there's no translation available in the current language
        queryset = get_translation_queryset(self)
        if queryset:
            return queryset[0]
        # If there's no translation available at all
        return None


class Enquiry(TransModelMixin, models.Model):
    """
    An enquiry has start and end dates and a question with several answers.

    For translateable fields see the ``EnquiryTrans`` model.

    :creation_date: The DateTime when this enquiry was created.
    :created_by: The user who created this enquiry.
    :start_date: The start date of this enquiry.
    :end_date: The end date of this enquiry.
    :allow_anonymous: If ``True`` anonymous user can participate, otherwise
      users must be authenticated in order to vote.

    """
    creation_date = models.DateTimeField(
        auto_now_add=True,
        verbose_name=_('Creation Date'),
    )

    created_by = models.ForeignKey(
        'auth.User',
        verbose_name=_('Created by'),
        related_name='enquiries',
        blank=True, null=True,
    )

    start_date = models.DateTimeField(
        default=lambda: timezone.now(),
        verbose_name=_('Start'),
    )

    end_date = models.DateTimeField(
        default=lambda: timezone.now() + timezone.timedelta(days=7),
        verbose_name=_('End'),
        blank=True, null=True,
    )

    allow_anonymous = models.BooleanField(
        default=False,
        verbose_name=_('Allow anonymous votes'),
    )

    is_published = models.BooleanField(
        default=False,
        verbose_name=('Is published'),
    )

    class Meta:
        ordering = ('-start_date', )

    def __unicode__(self):
        translation = self.get_translation()
        if translation:
            return translation.question
        return 'not translated'

    def get_answers(self):
        """Returns a queryset of all answers to its question."""
        return [answer.get_translation() for answer
                in self.answers.all() if answer.get_translation()]

    def has_voted(self, request):
        """Returns True if the current User has already voted."""
        answer_pks = [answer.pk for answer in self.answers.all()
                      if answer.get_translation()]
        if request.user.is_authenticated():
            votes = Vote.objects.filter(
                models.Q(session_key=request.session.session_key) | models.Q(
                    user=request.user), answer__pk__in=answer_pks)
        else:
            votes = Vote.objects.filter(
                session_key=request.session.session_key,
                answer__pk__in=answer_pks)
        if votes:
            return True
        return False

    def is_active(self):
        """Returns True if enquiry's campaign is still running."""
        if (self.start_date < timezone.now()
                and (not self.end_date or self.end_date > timezone.now())):
            return True
        return False


class EnquiryTrans(models.Model):
    """
    Translateable fields for the ``Enquiry`` model.

    :question: The title of this category.
    :description: Additional text to describe the purpose of the poll
    :extra_info: Additional text that can be used for anything.

    """
    question = models.CharField(
        max_length=100,
        verbose_name=_('Question'),
    )

    description = models.TextField(
        verbose_name=_('Description'),
        blank=True,
    )

    extra_info = models.TextField(
        verbose_name=_('Extra info'),
        blank=True,
    )

    # Needed by simple-translation
    enquiry = models.ForeignKey(Enquiry, verbose_name=_('Enquiry'))
    language = models.CharField(
        max_length=2, verbose_name=_('Language'), choices=settings.LANGUAGES)

    def __unicode__(self):
        return self.question


class Answer(TransModelMixin, models.Model):
    """
    An answer belongs to an enquiry.

    For translateable fields see ``AnswerTrans`` model.

    :enquiry: FK to the enquiry this answer belongs to.

    """
    enquiry = models.ForeignKey(
        'enquiry.Enquiry',
        verbose_name=_('Enquiry'),
        related_name='answers',
    )

    def __unicode__(self):
        translation = self.get_translation()
        if translation:
            return translation.text
        return 'not translated'

    def get_vote_count(self):
        """Returns the number of votes for this answer."""
        return self.votes.all().count()


class AnswerTrans(models.Model):
    """
    Translateable fields for the ``Answer`` model.

    :text: The text of this answer.

    """
    text = models.CharField(
        max_length=100,
        verbose_name=_('Answer'),
    )

    # Needed by simple-translation
    answer = models.ForeignKey(Answer, verbose_name=_('Answer'))
    language = models.CharField(
        max_length=2, verbose_name=_('Language'), choices=settings.LANGUAGES)

    def __unicode__(self):
        return self.text


class Vote(models.Model):
    """
    A vote is being cast on an answer by a user.

    :answer: The answer this user has chosen.
    :user: The user who cast this vote. Can be None if the user was anonymous.
    :session_key: Stores the current session key if vote has been anonymous.

    """
    answer = models.ForeignKey(
        'enquiry.Answer',
        verbose_name=_('Answer'),
        related_name='votes',
    )

    user = models.ForeignKey(
        'auth.User',
        verbose_name=_('User'),
        related_name='votes',
        blank=True, null=True,
    )

    session_key = models.CharField(
        max_length=100,
        verbose_name=_('User'),
        blank=True,
    )

    def __unicode__(self):
        return '{0}'.format(self.answer)



"""Admin classes for the ``enquiry`` app."""
from django.contrib import admin

from simple_translation.admin import TranslationAdmin

from enquiry.models import Answer, Enquiry, Vote


class AnswerAdmin(TranslationAdmin):
    pass


class EnquiryAdmin(TranslationAdmin):
    list_display = [
        'question', 'start_date', 'end_date', 'is_published', 'languages', ]

    def question(self, obj):
       return obj.get_translation().question


admin.site.register(Answer, AnswerAdmin)
admin.site.register(Enquiry, EnquiryAdmin)
admin.site.register(Vote)



"""
This ``urls.py`` is only used when running the tests via ``runtests.py``.
As you know, every app must be hooked into yout main ``urls.py`` so that
you can actually reach the app's views (provided it has any views, of course).

"""
from django.conf import settings
from django.conf.urls.defaults import include, patterns, url
from django.contrib import admin
from django.contrib.staticfiles.urls import staticfiles_urlpatterns
from django.conf.urls.static import static


admin.autodiscover()

urlpatterns = static(settings.MEDIA_URL, document_root=settings.MEDIA_ROOT)
urlpatterns += staticfiles_urlpatterns()
urlpatterns += patterns(
    '',
    url(r'^admin/', include(admin.site.urls)),
    url(r'^', include('enquiry.urls')),
)



"""Template tags for the ``enquiry`` app."""
from django import template
from django.db.models import Q
from django.utils.timezone import now

from enquiry.models import Enquiry

register = template.Library()


@register.assignment_tag
def get_answers(enquiry, sort=False):
    """
    Returns the answers of a given enquiry.

    :param enquiry: The given Enquiry instance.
    :param sort: If true, the answers will be sorted by vote count descending.

    """
    answers = enquiry.get_answers()
    if sort:
        answers.sort(key=lambda a: a.answer.get_vote_count(), reverse=True)
    return answers


@register.inclusion_tag('enquiry/current_poll.html', takes_context=True)
def render_current_poll(context):
    """Template tag to render one of the current active polls."""
    enquiries = Enquiry.objects.filter(
        Q(is_published=True),
        Q(start_date__lt=now()),
        Q(end_date__gt=now()) | Q(end_date__isnull=True))
    if enquiries:
        return {
            'enquiry': enquiries[0],
            'has_voted': enquiries[0].has_voted(context['request'])
        }
    return ''


@register.assignment_tag
def get_current_poll():
    """Returns the current active poll."""
    enquiries = Enquiry.objects.filter(
        Q(is_published=True),
        Q(start_date__lt=now()),
        Q(end_date__gt=now()) | Q(end_date__isnull=True))
    if enquiries:
        return enquiries[0]
    return None



"""Registration of models of the ``enquiry`` app."""
from simple_translation.translation_pool import translation_pool

from enquiry.models import (
    Answer,
    AnswerTrans,
    Enquiry,
    EnquiryTrans,
)


translation_pool.register_translation(Answer, AnswerTrans)
translation_pool.register_translation(Enquiry, EnquiryTrans)
